# -*- coding: utf-8 -*-

import os
import psycopg2
import psycopg2.extras

from pirogue.utils import table_parts, select_columns, insert_command, update_command
from pirogue.information_schema import reference_columns, primary_key, default_value
from pirogue.exceptions import TableHasNoPrimaryKey


class SingleInheritance:
    """
    Creates a join view with associated triggers to edit for a single inheritance.
    """

    def __init__(self, parent_table: str, child_table: str,
                 pg_service: str = None,
                 view_schema: str = None,
                 view_name: str = None,
                 pkey_default_value: bool = False):
        """
        Produces the SQL code of the join table and triggers

        Parameters
        ----------
        pg_service
            if not given, it is determined using environment variable PGSERVICE
        parent_table
            the parent table, can be schema specified
        child_table
            the child table, can be schema specified
        view_schema
            the schema where the view will written to
        view_name
            the name of the created view, defaults to vw_{parent_table}_{join_table}
        pkey_default_value
            the primary key column of the view will have a default value according to the child primary key table
        """

        if pg_service is None:
            pg_service = os.getenv('PGSERVICE')
        self.conn = psycopg2.connect("service={0}".format(pg_service))
        self.cursor = self.conn.cursor()

        self.pkey_default_value = pkey_default_value

        (self.parent_schema, self.parent_table) = table_parts(parent_table)
        (self.child_schema, self.child_table) = table_parts(child_table)

        if view_schema is None:
            if self.parent_schema != self.child_schema:
                raise ValueError('Destination schema cannot be guessed if different on sources tables.')
            else:
                self.view_schema = self.parent_schema
        else:
            self.view_schema = view_schema

        self.view_name = view_name or "vw_{pt}_{ct}".format(pt=self.parent_table, ct=self.child_table)

        (self.ref_parent_key, parent_referenced_key) = reference_columns(self.cursor,
                                                                         self.child_schema, self.child_table,
                                                                         self.parent_schema, self.parent_table)
        try:
            self.child_pkey = primary_key(self.cursor, self.child_schema, self.child_table)
        except TableHasNoPrimaryKey:
            self.child_pkey = self.ref_parent_key
        self.parent_pkey = primary_key(self.cursor, self.parent_schema, self.parent_table)

        assert self.parent_pkey == parent_referenced_key

    def create(self) -> bool:
        """
        Creates the merge view on the specified service
        Returns True in case of success
        """
        success = True
        for sql in [self.__view(),
                    self.__insert_trigger(),
                    self.__update_trigger(),
                    self.__delete_trigger(),
                    self.__extras()
                    ]:
            try:
                if sql:
                    self.cursor.execute(sql)
            except psycopg2.Error as e:
                success = False
                print("*** Failing:\n{}\n***".format(sql))
                raise e
        self.conn.commit()
        self.conn.close()
        return success

    def __view(self) -> str:
        """
        Create the SQL code for the view
        :return: the SQL code
        """
        sql = """
CREATE OR REPLACE VIEW {vs}.{vn} AS SELECT
  {child_cols},
  {parent_cols}
  FROM {cs}.{ct}
  LEFT JOIN {ps}.{pt} ON {pt}.{prk} = {ct}.{rpk};
""".format(vs=self.view_schema,
           vn=self.view_name,
           parent_cols=select_columns(self.cursor, self.parent_schema, self.parent_table, table_alias=self.parent_table, remove_pkey=True),
           child_cols=select_columns(self.cursor, self.child_schema, self.child_table, table_alias=self.child_table),
           cs=self.child_schema,
           ct=self.child_table,
           ps=self.parent_schema,
           pt=self.parent_table,
           rpk=self.ref_parent_key,
           prk=self.parent_pkey)

        return sql

    def __insert_trigger(self) -> str:
        """

        :return:
        """
        sql = """
-- INSERT TRIGGER
CREATE OR REPLACE FUNCTION {vs}.ft_{vn}_insert() RETURNS trigger AS
$BODY$
BEGIN
{insert_parent}

{insert_child}
RETURN NEW;
END;
$BODY$
LANGUAGE plpgsql;

DROP TRIGGER IF EXISTS tr_{vn}_on_insert ON {vs}.{vn};

CREATE TRIGGER tr_{vn}_on_insert
  INSTEAD OF INSERT ON {vs}.{vn}
  FOR EACH ROW EXECUTE PROCEDURE {vs}.ft_{vn}_insert();
""".format(vs=self.view_schema,
           vn=self.view_name,
           insert_parent=insert_command(self.cursor, self.parent_schema, self.parent_table,
                                        remove_pkey=False,
                                        coalesce_pkey_default=True,
                                        remap_columns={self.parent_pkey: self.ref_parent_key},
                                        returning='{ppk} INTO NEW.{prk}'.format(ppk=self.parent_pkey, prk=self.ref_parent_key)),
           insert_child=insert_command(self.cursor, self.child_schema, self.child_table,
                                       remove_pkey=False,
                                       pkey=self.child_pkey)
           )
        return sql

    def __update_trigger(self):
        sql = """
-- UPDATE TRIGGER
CREATE OR REPLACE FUNCTION {vs}.ft_{vn}_update() RETURNS trigger AS
$BODY$
BEGIN
{update_master}

{update_child}
RETURN NEW;
END;
$BODY$
LANGUAGE plpgsql;

DROP TRIGGER IF EXISTS tr_{vn}_on_update ON {vs}.{vn};

CREATE TRIGGER tr_{vn}_on_update
  INSTEAD OF UPDATE ON {vs}.{vn}
  FOR EACH ROW EXECUTE PROCEDURE {vs}.ft_{vn}_update();
""".format(vs=self.view_schema,
           vn=self.view_name,
           update_master=update_command(self.cursor, self.parent_schema, self.parent_table,
                                        remap_columns={self.parent_pkey: self.ref_parent_key}),
           update_child=update_command(self.cursor, self.child_schema, self.child_table, pkey=self.child_pkey, remove_pkey=False)
           )
        return sql

    def __delete_trigger(self):
        sql = """
CREATE OR REPLACE FUNCTION {vs}.ft_{vn}_delete() RETURNS trigger AS
$BODY$
BEGIN
  DELETE FROM {cs}.{ct} WHERE {rpk} = OLD.{rpk};
  DELETE FROM {ps}.{pt} WHERE {ppk} = OLD.{rpk};
RETURN NULL;
END;
$BODY$
LANGUAGE plpgsql;

DROP TRIGGER IF EXISTS tr_{vn}_on_delete ON {vs}.{vn};

CREATE TRIGGER tr_{vn}_on_delete
  INSTEAD OF DELETE ON {vs}.{vn}
  FOR EACH ROW EXECUTE PROCEDURE {vs}.ft_{vn}_delete();
""".format(vs=self.view_schema,
           vn=self.view_name,
           cs=self.child_schema,
           ct=self.child_table,
           rpk=self.ref_parent_key,
           ppk=self.parent_pkey,
           ps=self.parent_schema,
           pt=self.parent_table)
        return sql

    def __extras(self):
        sql = ''
        if self.pkey_default_value:
            sql += "ALTER VIEW {vs}.{vn} ALTER {rpk} SET DEFAULT {dv};"\
                .format(vs=self.view_schema,
                        vn=self.view_name,
                        rpk=self.child_pkey,
                        dv=default_value(self.cursor, self.child_schema, self.child_table, self.child_pkey))
        return sql


#!/usr/bin/env python3

import argparse
import os
import yaml

from pirogue.single_inheritance import SingleInheritance
from pirogue.multiple_inheritance import MultipleInheritance
from pirogue.simple_joins import SimpleJoins


def main():

    # create the top-level parser
    parser = argparse.ArgumentParser()
    parser.add_argument("-v", "--version", help="print the version and exit", action='store_true')

    subparsers = parser.add_subparsers(title='commands', description='pirogue command', dest='command')

    # single inheritance view
    single_inheritance_parser = subparsers.add_parser('single_inheritance', help='create a single inheritance view')
    single_inheritance_parser.add_argument('parent_table')
    single_inheritance_parser.add_argument('child_table')
    single_inheritance_parser.add_argument('-s', '--view-schema', help='schema for the created join view')
    single_inheritance_parser.add_argument('-v', '--view-name', help='name for the created join view')
    single_inheritance_parser.add_argument('-d', '--pkey-default-value', action='store_true',
                                           help='The primary key column of the view will have a default value'
                                                ' according to the child primary key table')
    single_inheritance_parser.add_argument('-p', '--pg_service', help='postgres service')


    # multiple inheritance view
    multiple_inheritance_parser = subparsers.add_parser('multiple_inheritance', help='create a multiple inheritance view')
    multiple_inheritance_parser.add_argument('definition_file', help='YAML definition of the merge view', type=argparse.FileType('r'))
    multiple_inheritance_parser.add_argument('-j', '--create-joins', action='store_true',
                                             help='Create simple join view for all joined tables.')
    multiple_inheritance_parser.add_argument('-d', '--drop', action='store_true',
                                             help='Drop existing views, type and triggers.')
    multiple_inheritance_parser.add_argument('-v', '--var', nargs=3,
                                             help='Assign variable for running SQL deltas. '
                                                  'Format is: (string|float|int) name value. ',
                                             action='append', default=[])
    multiple_inheritance_parser.add_argument('-p', '--pg_service', help='postgres service')


    # multiple inheritance view
    simple_joins = subparsers.add_parser('simple_joins', help='create a view for simple joins without any editing capability')
    simple_joins.add_argument('definition_file', help='YAML definition of the merge view', type=argparse.FileType('r'))
    simple_joins.add_argument('-p', '--pg_service', help='postgres service')

    args = parser.parse_args()

    # print the version and exit
    if args.version:
        import pkg_resources
        print('pirogue version: {}'.format(pkg_resources.get_distribution('pirogue').version))
        parser.exit()

    # if no command is passed, print the help and exit
    if not args.command:
        parser.print_help()
        parser.exit()

    exit_val = 0

    if args.pg_service:
        pg_service = args.pg_service
    else:
        pg_service = os.getenv('PGSERVICE')

    if args.command == 'single_inheritance':
        success = SingleInheritance(parent_table=args.parent_table,
                                    child_table=args.child_table,
                                    pg_service=pg_service,
                                    view_schema=args.view_schema,
                                    view_name=args.view_name,
                                    pkey_default_value=args.pkey_default_value).create()
        if not success:
            exit_val = 1

    elif args.command == 'multiple_inheritance':
        yaml_definition = yaml.safe_load(args.definition_file)
        variables = {}
        for v in args.var or ():
            if v[0] == 'float':
                variables[v[1]] = float(v[2])
            elif v[0] == 'int':
                variables[v[1]] = int(v[2])
            else:
                variables[v[1]] = v[2]
        MultipleInheritance(yaml_definition,
                            variables=variables,
                            create_joins=args.create_joins,
                            pg_service=pg_service).create()

    elif args.command == 'simple_joins':
        yaml_definition = yaml.safe_load(args.definition_file)
        SimpleJoins(yaml_definition, pg_service=pg_service).create()


    exit(exit_val)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

import os
import psycopg2
import psycopg2.extras

from pirogue.utils import table_parts, select_columns
from pirogue.information_schema import reference_columns, primary_key
from pirogue.exceptions import InvalidDefinition, TableHasNoPrimaryKey, NoReferenceFound


class SimpleJoins:
    """
    Creates a view made of simple joins, without any edit triggers.
    """
    def __init__(self,
                 definition: dict,
                 pg_service: str = None):
        """
        Produces the SQL code of the join table and triggers

        Parameters
        ----------
        definition
            the YAML definition of the multiple inheritance
        pg_service
            if not given, it is determined using environment variable PGSERVICE
        """

        # check definition validity
        for key in definition.keys():
            if key not in ('table', 'view_schema','joins', 'pkey', 'view_name'):
                raise InvalidDefinition('key {k} is not a valid'.format(k=key))
        # check joins validity
        for alias, table_def in definition['joins'].items():
            for key in table_def.keys():
                if key not in ('table', 'short_alias',
                               'fkey', 'prefix',
                               'skip_columns', 'remap_columns'):
                    raise InvalidDefinition('in join {a} key "{k}" is not valid'.format(a=alias, k=key))

        if pg_service is None:
            pg_service = os.getenv('PGSERVICE')
        self.conn = psycopg2.connect("service={0}".format(pg_service))
        self.cursor = self.conn.cursor()

        (self.parent_schema, self.parent_table) = table_parts(definition['table'])
        self.view_schema = definition.get('view_schema', self.parent_schema)
        self.view_name = definition.get('view_name', "vw_{pt}".format(pt=self.parent_table))

        try:
            self.parent_pkey = primary_key(self.cursor, self.parent_schema, self.parent_table)
        except TableHasNoPrimaryKey:
            self.parent_pkey = definition['pkey']

        class Table:
            def __init__(self):
                self.schema_name = None
                self.table_name = None
                self.pkey = None
                self.ref_parent_key = None
                self.parent_referenced_key = None
                self.skip_columns = None
                self.remap_columns = None
                self.prefix = None

        self.child_tables = {}
        for alias, table_def in definition['joins'].items():
            child = Table()
            (child.schema_name, child.table_name) = table_parts(table_def['table'])
            child.pkey = primary_key(self.cursor, child.schema_name, child.table_name)
            try:
                (child.parent_referenced_key, child.ref_parent_key) = reference_columns(self.cursor,
                                                                                        self.parent_schema,
                                                                                        self.parent_table,
                                                                                        child.schema_name,
                                                                                        child.table_name)
                assert child.pkey == child.ref_parent_key
            except NoReferenceFound:
                child.parent_referenced_key = table_def['fkey']

            child.skip_columns = table_def.get('skip_columns', {})
            child.remap_columns = table_def.get('remap_columns', {})
            child.prefix = table_def.get('prefix', None)
            self.child_tables[alias] = child

    def create(self) -> bool:
        """
        Creates the merge view on the specified service
        Returns True in case of success
        """
        sql = self.__view()
        success = True
        try:
            self.cursor.execute(sql)
        except psycopg2.Error as e:
            success = False
            print("*** Failing:\n{}\n***".format(sql))
            raise e
        self.conn.commit()
        self.conn.close()
        return success

    def __view(self) -> str:
        """
        Create the SQL code for the view
        :return: the SQL code
        """
        sql = """
CREATE OR REPLACE VIEW {vs}.{vn} AS SELECT
  {parent_cols},
  {child_cols}
  FROM {ps}.{pt}
  {joins};
""".format(vs=self.view_schema,
           vn=self.view_name,
           parent_cols=select_columns(self.cursor, self.parent_schema, self.parent_table, table_alias=self.parent_table, remove_pkey=False),
           child_cols=',\n  '.join([select_columns(self.cursor, child_def.schema_name, child_def.table_name,
                                                   table_alias=alias, remap_columns=child_def.remap_columns,
                                                   prefix=child_def.prefix)
                                    for alias, child_def in self.child_tables.items()]),
           ps=self.parent_schema,
           pt=self.parent_table,
           joins='\n  '.join(["LEFT JOIN {cs}.{ct} {alias} ON {alias}.{cpk} = {pt}.{rpk}".format(cs=child.schema_name,
                                                                                                 ct=child.table_name,
                                                                                                 alias=alias,
                                                                                                 cpk=child.pkey,
                                                                                                 pt=self.parent_table,
                                                                                                 rpk=child.parent_referenced_key)
                            for alias, child in self.child_tables.items()])
           )

        return sql


from setuptools import setup
import sys

python_min_version = (3, 6)

if sys.version_info < python_min_version:
    sys.exit('Pirogue requires at least Python version {vmaj}.{vmin}.\n'
             'You are currently running this installation with\n\n{curver}'.format(
        vmaj=python_min_version[0],
        vmin=python_min_version[1],
        curver=sys.version))

setup(
    name='pirogue',
    packages=[
        'pirogue',
        'scripts'
    ],
    entry_points={
        'console_scripts': [
            'pirogue = scripts.pirogue:main'
        ]
    },
    version='1.0.4',
    description='PostgreSQL view generator',
    author='Denis Rouzaud',
    author_email='denis.rouzaud@gmail.com',
    url='https://github.com/opengisch/pirogue',
    download_url='https://github.com/opengisch/pirogue/archive/1.0.4.tar.gz',
    keywords=['postgres'],
    classifiers=[
        'Topic :: Database',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Intended Audience :: System Administrators',
        'Intended Audience :: Information Technology',
        'Development Status :: 3 - Alpha'
    ],
    install_requires=[
        'psycopg2-binary>=2.7.3'
    ],
    python_requires=">={vmaj}.{vmin}".format(vmaj=python_min_version[0], vmin=python_min_version[1]),
)
